/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 *    PreferenceStructurePanel.java
 *    Copyright (C) 2008 Jens Christian Hhn
 *
 *    (based upon)  BoundaryPanel.java
 *    Copyright (C) 2002 University of Waikato, Hamilton, New Zealand
 *
 */


package weka.gui.extensions;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Locale;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.ToolTipManager;

import weka.classifiers.Classifier;
import weka.classifiers.PreferenceStructureHandler;
import weka.core.Instance;
import weka.core.Instances;

import com.sun.image.codec.jpeg.JPEGCodec;
import com.sun.image.codec.jpeg.JPEGEncodeParam;
import com.sun.image.codec.jpeg.JPEGImageEncoder;

/**
 * PreferenceStructurePanel. A class to generate a visualization of a 
 * fuzzy preference structure.
 * 
 * Date created: 08/07/2008
 * @author Jens Christian Hhn (huehn@gmx.net)
 * @version $Revision: 1.0 $
 * @see JPanel
 */
public class PreferenceStructurePanel extends JPanel {

  /** for serialization */
  private static final long serialVersionUID = -8499445518744770458L;

  /** training data */
  protected Instances m_trainingData;

  /** The whole dataset without one instance which preference structure shall be visualized */
  protected Instances m_LOOData;

  /** The instance for which the preference structure shall be visualized */
  protected Instance m_Instance;

  /** distribution classifier to use */
  protected Classifier m_classifier;

  /** Flag marking that a classifier was built */
  protected boolean m_classifierWasBuilt = false;

  /** Switch between colored and grayscale output */
  protected boolean m_Grayscale = false;

  // attributes for visualizing on
  protected int m_xAttribute;

  protected int m_yAttribute;

  // min, max and ranges of these attributes
  protected double m_minX;

  protected double m_minY;

  protected double m_maxX;

  protected double m_maxY;

  protected String m_latexCode = null;
  
  // pixel width and height in terms of attribute values
  protected double m_pixHeight;

  protected double m_pixWidth;

  /** used for offscreen drawing */
  protected Image m_osi = null;



  /** listeners to be notified when plot is complete */
  private Vector m_listeners = new Vector();

  /**
   * small inner class for rendering the bitmap on to
   */
  private class PlotPanel extends JPanel {

    /** for serialization */
    private static final long serialVersionUID = 743629498352235060L;

    public PlotPanel() {
      this.setToolTipText("");
    }

    public void paintComponent(Graphics g) {
      super.paintComponent(g);
      if (m_osi != null) {
	g.drawImage(m_osi, 0, 0, this);
      }
    }

    public String getToolTipText(MouseEvent event) {
      return null;
    }

  }

  /** the actual plotting area */
  private PlotPanel m_plotPanel = new PlotPanel();

  /** thread for running the plotting operation in */
  private Thread m_plotThread = null;

  /** Stop the plotting thread */
  protected boolean m_stopPlotting = false;

  /** Stop any replotting threads */
  protected boolean m_stopReplotting = false;


  /** plot the training data */
  protected boolean m_plotTrainingData = true;

  /**
   * Creates a new <code>PreferenceStructurePanel</code> instance.
   * 
   * @param panelWidth
   *                the width in pixels of the panel
   * @param panelHeight
   *                the height in pixels of the panel
   */
  public PreferenceStructurePanel(int panelWidth, int panelHeight) {
    ToolTipManager.sharedInstance().setDismissDelay(Integer.MAX_VALUE);

    setLayout(new BorderLayout());
    m_plotPanel.setMinimumSize(new Dimension(panelWidth, panelHeight));
    m_plotPanel.setPreferredSize(new Dimension(panelWidth, panelHeight));
    m_plotPanel.setMaximumSize(new Dimension(panelWidth, panelHeight));
    add(m_plotPanel, BorderLayout.CENTER);
    setPreferredSize(m_plotPanel.getPreferredSize());
    setMaximumSize(m_plotPanel.getMaximumSize());
    setMinimumSize(m_plotPanel.getMinimumSize());

  }

  /**
   * Set up the off screen bitmap for rendering to
   */
  protected void initialize() {
    if (m_plotPanel.getWidth() > 0 && m_plotPanel.getHeight() > 0) {
      m_osi = m_plotPanel.createImage(m_plotPanel.getWidth(), m_plotPanel
	  .getHeight());
      Graphics m = m_osi.getGraphics();
      m.setColor(Color.white);
      m.fillRect(0, 0, m_plotPanel.getWidth(), m_plotPanel.getHeight());
    }
  }

  /**
   * Stop the plotting thread
   */
  public void stopPlotting() {
    m_stopPlotting = true;
    try {
      m_plotThread.join(100);
    } catch (Exception e) {
    }
    ;
  }

  /**
   * Start the plotting thread
   * 
   * @exception Exception
   *                    if an error occurs
   */
  public void start() throws Exception {

    m_stopReplotting = true;
    if (m_trainingData == null) {
      throw new Exception("No training data set (PreferenceStructurePanel)");
    }
    if (m_classifier == null) {
      throw new Exception("No classifier set (PreferenceStructurePanel)");
    }
    
    if (!(m_classifier instanceof PreferenceStructureHandler)) {
      throw new Exception("A Fuzzy Preference Structure Classifier has to be set!");
    }

    m_classifier.buildClassifier(m_LOOData);
    m_classifierWasBuilt = true;
    m_classifier.classifyInstance(m_Instance);

    startPlotThread();
  }

  // Thread for main plotting operation
  protected class PlotThread extends Thread {
    double[] m_weightingAttsValues;

    boolean[] m_attsToWeightOn;

    double[] m_vals;

    double[] m_dist;

    Instance m_predInst;

    public void run() {

      m_stopPlotting = false;
      try {
	initialize();
	repaint();

	// train the classifier
	m_classifier.buildClassifier(m_trainingData);

	// build DataGenerator
	m_attsToWeightOn = new boolean[m_trainingData.numAttributes()];
	m_attsToWeightOn[m_xAttribute] = true;
	m_attsToWeightOn[m_yAttribute] = true;

	// generate samples
	m_weightingAttsValues = new double[m_attsToWeightOn.length];
	m_vals = new double[m_trainingData.numAttributes()];
	m_predInst = new Instance(1.0, m_vals);
	m_predInst.setDataset(m_trainingData);

	update();

	if (m_plotTrainingData) {
	  plotTrainingData();
	}

      } catch (Exception ex) {
	ex.printStackTrace();
	JOptionPane.showMessageDialog(null, "Error while plotting: \""
	    + ex.getMessage() + "\"");
      } finally {
	m_plotThread = null;
	// notify any listeners that we are finished
	Vector l;
	ActionEvent e = new ActionEvent(this, 0, "");
	synchronized (this) {
	  l = (Vector) m_listeners.clone();
	}
	for (int i = 0; i < l.size(); i++) {
	  ActionListener al = (ActionListener) l.elementAt(i);
	  al.actionPerformed(e);
	}
      }
    }

  }

  /**
   * Render the training points on-screen.
   */
  public void plotTrainingData() {
    
    String latexCode = new String("");

    Graphics2D osg = (Graphics2D) m_osi.getGraphics();
    Graphics g = m_plotPanel.getGraphics();

    osg.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
	RenderingHints.VALUE_ANTIALIAS_ON);

    osg.setColor(Color.gray);

    int boxDistance = (this.getSize().width < this.getSize().height ? this
	.getSize().width : this.getSize().height)
	/ (m_trainingData.numClasses() + (this.getSize().width < this.getSize().height ? 1
	    : 0));
    int maxBoxLength = 9
	* (this.getSize().width < this.getSize().height ? this.getSize().width
	    : this.getSize().height)
	/ (m_trainingData.numClasses() + (this.getSize().width < this.getSize().height ? 1
	    : 0)) / 10;

    double[][][] relations = null;
    double[] dist = null;
    if (m_classifierWasBuilt && m_classifier != null
	&& m_classifier instanceof PreferenceStructureHandler) {

      
      
      if (m_Grayscale){
	latexCode += "\\definecolor{PFVCOLORlightgray}{gray}{0.8}\r\n";
	latexCode += "\\definecolor{PFVCOLORmediumgray}{gray}{0.6}\r\n";
	latexCode += "\\definecolor{PFVCOLORdarkgray}{gray}{0.4}\r\n";
      }
      
      latexCode += "% Do not forget to include packages tabularx and rotating!\r\n";
      latexCode += "\\renewcommand\\tabularxcolumn[1]{m{#1}}\r\n";
      latexCode += "\\newcolumntype{Y}{>{\\centering\\arraybackslash}X}";
      
      
      latexCode += "\\begin{figure}\r\n";
      latexCode += "\\centering\r\n";
      
      
      DecimalFormat formatInt = new DecimalFormat("#####.#######");
      formatInt.setDecimalFormatSymbols(new DecimalFormatSymbols(Locale.US));
      latexCode += "\\setlength{\\unitlength}{" + formatInt.format(1.0/(1.3*(m_trainingData.numClasses()+2.5)*102.0)) + "\\textwidth}\r\n";
      latexCode += "\\begin{tabularx}{\\textwidth}{m{" + formatInt.format(1.5/(1.3*(m_trainingData.numClasses()+2.5))) + "\\textwidth}|";
      
      
      for (int i = 0; i < m_trainingData.numClasses(); i++) {
	latexCode += "Y";
      }
      latexCode += "Y} \r\n";
      
      latexCode += "& ";
      for (int i = 0; i < m_trainingData.numClasses(); i++) {


	latexCode += m_trainingData.classAttribute().value(i);
	if (i < m_trainingData.numClasses() - 1) {
	  latexCode += " & ";
	} else {
	  latexCode += " &Score\\\\\\hline\\\\[-10pt]\r\n";
	}
      }
      
      latexCode += "%& ";
      for (int i = 0; i < m_trainingData.numClasses(); i++) {
	latexCode += "\\begin{sideways}" + m_trainingData.classAttribute().value(i)+ "\\end{sideways}";
	if (i < m_trainingData.numClasses() - 1) {
	  latexCode += " & ";
	} else {
	  latexCode += " &\\begin{sideways}Score\\end{sideways}\\\\\\hline\\\\[-10pt]\r\n";
	}
      }
      
      
      
      

      try {
	relations = ((PreferenceStructureHandler) m_classifier)
	    .relationsForInstance(m_Instance);
	dist = m_classifier.distributionForInstance(m_Instance);
      } catch (Exception e) {
	e.printStackTrace();
      }

      for (int i = 0; i < m_trainingData.numClasses(); i++) {
	latexCode += m_trainingData.classAttribute().value(i) + " & ";

	for (int j = 0; j < m_trainingData.numClasses(); j++) {
	  if (relations != null) {
	    if (i != j) {
	      double preference = relations[0][i][j];
	      double conflict = relations[1][i][j];
	      double ignorance = relations[2][i][j];
	      double oppositePreference = relations[0][j][i];

	      if (ignorance < 1.0) {
		preference /= (1 - ignorance);
		conflict /= (1 - ignorance);
		oppositePreference /= (1 - ignorance);
	      }

	      int PAD = 8;

	      if (ignorance < 1 && preference > 0) {
		if (m_Grayscale)
		  osg.setColor(Color.lightGray);
		else
		  osg.setColor(Color.green);
		osg.fillRect(boxDistance * j + PAD, boxDistance * i + PAD,
		    (int) Math.floor(maxBoxLength * (1 - ignorance)) - PAD / 2,
		    (int) Math
			.ceil((maxBoxLength * (1 - ignorance) * preference))
			- PAD / 2 + 1);
	      }

	      if (ignorance < 1 && conflict > 0) {
		if (m_Grayscale)
		  osg.setColor(Color.gray);
		else
		  osg.setColor(Color.yellow);
		int yellowOffset = 1;
		if (preference > 0)
		  yellowOffset = -2;
		osg.fillRect(boxDistance * j + PAD, boxDistance
		    * i
		    + PAD
		    + (int) Math.floor(maxBoxLength * (1 - ignorance)
			* preference) + yellowOffset,
		    (int) (maxBoxLength * (1 - ignorance)) - PAD / 2,
		    (int) Math.ceil(maxBoxLength * (1 - ignorance) * conflict)
			- PAD / 2 + 2);
	      }

	      if (ignorance < 1 && oppositePreference > 0) {
		if (m_Grayscale)
		  osg.setColor(Color.darkGray);
		else
		  osg.setColor(Color.red);
		osg.fillRect(boxDistance * j + PAD, boxDistance
		    * i
		    + PAD
		    + (int) Math.floor(maxBoxLength * (1 - ignorance)
			* preference)
		    + (int) Math.floor(maxBoxLength * (1 - ignorance)
			* conflict), (int) Math.ceil(maxBoxLength
		    * (1 - ignorance))
		    - PAD / 2,
		    (int) (maxBoxLength * (1 - ignorance) * oppositePreference)
			- PAD / 2 + 1);
	      }

	      osg.setColor(Color.black);
	      if (ignorance < 1)
		osg.drawRect(boxDistance * j + PAD, boxDistance * i + PAD,
		    (int) (maxBoxLength * (1 - ignorance)) - PAD / 2,
		    (int) (maxBoxLength * (1 - ignorance)) - PAD / 2);

	      latexCode += "\\begin{picture}(102,102)(0,0)";

	      if (ignorance < 1)
		latexCode += "\\put(0,0){\\rule{"
		    + ((1 - ignorance) * 100.0 + 2) + "\\unitlength}{"
		    + ((1 - ignorance) * 100.0 + 2) + "\\unitlength}}\r\n";

	      if (oppositePreference > 0) {
		if (!m_Grayscale)
		  latexCode += "\\color{red}\r\n";
		else
		  latexCode += "\\color{PFVCOLORdarkgray}\r\n";
		latexCode += "\\put(1,1){\\rule{" + ((1 - ignorance) * 100.0)
		    + "\\unitlength}{" + (1 - ignorance) * oppositePreference
		    * 100.0 + "\\unitlength}}\r\n";
	      }

	      if (conflict > 0) {
		if (!m_Grayscale)
		  latexCode += "\\color{yellow}\r\n";
		else
		  latexCode += "\\color{PFVCOLORmediumgray}\r\n";
		latexCode += "\\put(1,"
		    + (1 + (1 - ignorance) * oppositePreference * 100.0)
		    + "){\\rule{" + (1 - ignorance) * 100.0 + "\\unitlength}{"
		    + (1 - ignorance) * conflict * 100.0 + "\\unitlength}}\r\n";
	      }

	      if (preference > 0) {
		if (!m_Grayscale)
		  latexCode += "\\color{green}\r\n";
		else
		  latexCode += "\\color{PFVCOLORlightgray}\r\n";
		latexCode += "\\put(1,"
		    + (1 + (1 - ignorance) * oppositePreference * 100.0 + (1 - ignorance)
			* conflict * 100.0) + "){\\rule{" + (1 - ignorance)
		    * 100.0 + "\\unitlength}{" + (1 - ignorance) * preference
		    * 100.0 + "\\unitlength}}\r\n";
	      }

	      latexCode += "\\end{picture}\r\n";

	    }
	    if (j < m_trainingData.numClasses() - 1)
	      latexCode += " & ";
	  }
	}

	osg.drawString(
	    new Double(Math.round(dist[i] * 100) / 100.0).toString(),
	    boxDistance * m_trainingData.numClasses() + maxBoxLength / 2,
	    boxDistance * i + maxBoxLength / 2 + 5);

	latexCode +=  " & " + new Double(Math.round(dist[i] * 100) / 100.0).toString() + "\\\\\r\n";
      }

      latexCode += "\\end{tabularx}";
      
      latexCode += "\\caption{Preference Structures}\r\n";
      latexCode += "\\label{fig:PFS}\r\n";
      latexCode += "\\end{figure}\r\n";
      
      
      
    }

    g.drawImage(m_osi, 0, 0, m_plotPanel);
    
    m_latexCode =  latexCode;
  }

  /**
   * Update the rendered image.
   */
  private void update() {
    Graphics g = m_plotPanel.getGraphics();
    g.drawImage(m_osi, 0, 0, m_plotPanel);
  }

  /**
   * Set the training data to use
   * 
   * @param trainingData
   *                the training data
   * @exception Exception
   *                    if an error occurs
   */
  public void setTrainingData(Instances trainingData) throws Exception {

    m_trainingData = trainingData;
    if (m_trainingData.classIndex() < 0) {
      throw new Exception("No class attribute set (PreferenceStructurePanel)");
    }
  }



  /**
   * Register a listener to be notified when plotting completes
   * 
   * @param newListener
   *                the listener to add
   */
  public void addActionListener(ActionListener newListener) {
    m_listeners.add(newListener);
  }

  /**
   * Remove a listener
   * 
   * @param removeListener
   *                the listener to remove
   */
  public void removeActionListener(ActionListener removeListener) {
    m_listeners.removeElement(removeListener);
  }

  /**
   * Set the classifier to use.
   * 
   * @param classifier
   *                the classifier to use
   */
  public void setClassifier(Classifier classifier) {
    m_classifier = classifier;
  }

  /**
   * Set the x attribute index
   * 
   * @param xatt
   *                index of the attribute to use on the x axis
   * @exception Exception
   *                    if an error occurs
   */
  public void setXAttribute(int xatt) throws Exception {
    if (m_trainingData == null) {
      throw new Exception("No training data set (PreferenceStructurePanel)");
    }
    if (xatt < 0 || xatt > m_trainingData.numAttributes()) {
      throw new Exception("X attribute out of range (PreferenceStructurePanel)");
    }
    if (m_trainingData.attribute(xatt).isNominal()) {
      throw new Exception("Visualization dimensions must be numeric "
	  + "(PreferenceStructurePanel)");
    }

    m_xAttribute = xatt;
  }

  /**
   * Set the y attribute index
   * 
   * @param yatt
   *                index of the attribute to use on the y axis
   * @exception Exception
   *                    if an error occurs
   */
  public void setYAttribute(int yatt) throws Exception {
    if (m_trainingData == null) {
      throw new Exception("No training data set (PreferenceStructurePanel)");
    }
    if (yatt < 0 || yatt > m_trainingData.numAttributes()) {
      throw new Exception("X attribute out of range (PreferenceStructurePanel)");
    }
    if (m_trainingData.attribute(yatt).isNominal()) {
      throw new Exception("Visualization dimensions must be numeric "
	  + "(PreferenceStructurePanel)");
    }
    m_yAttribute = yatt;
  }

  /**
   * Set whether to superimpose the training data plot
   * 
   * @param pg
   *                a <code>boolean</code> value
   */
  public void setPlotTrainingData(boolean pg) {
    m_plotTrainingData = pg;
  }

  /**
   * Returns true if training data is to be superimposed
   * 
   * @return a <code>boolean</code> value
   */
  public boolean getPlotTrainingData() {
    return m_plotTrainingData;
  }

  /**
   * Quickly replot the display using cached probability estimates
   */
  public void replot() {
    m_stopReplotting = true;

    // wait 300 ms to give any other replot threads a chance to halt
    try {
      Thread.sleep(300);
    } catch (Exception ex) {
    }

    final Thread replotThread = new Thread() {
      public void run() {
	m_stopReplotting = false;

	update();
	if (m_plotTrainingData) {
	  plotTrainingData();
	}
      }
    };

    replotThread.start();
  }

  protected void saveImage(String fileName) {
    try {
      BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(
	  fileName));

      JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(out);
      BufferedImage bi = new BufferedImage(this.getSize().width,
	  this.getSize().height, BufferedImage.TYPE_INT_RGB);

      Graphics2D gr2 = bi.createGraphics();
      gr2.drawImage(m_osi, 0, 0, this.getSize().width, this.getSize().height,
	  null);

      JPEGEncodeParam param = encoder.getDefaultJPEGEncodeParam(bi);
      param.setQuality(1.0f, false);
      encoder.setJPEGEncodeParam(param);
      encoder.encode(bi);
      out.flush();
      out.close();
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  /**
   * Starts the plotting thread. Will also create it if necessary.
   */
  public void startPlotThread() {
    if (m_plotThread == null) {
      m_plotThread = new PlotThread();
      m_plotThread.setPriority(Thread.MIN_PRIORITY);
      m_plotThread.start();
    }
  }

  /**
   * Adds a mouse listener.
   */
  public void addMouseListener(MouseListener l) {
    m_plotPanel.addMouseListener(l);
  }

}
