/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *   PreferenceStructureVisualizer.java
 *    Copyright (C) 2008 Jens Christian Hhn
 *    
 *   (based upon) BoundaryVisualizer.java
 *   Copyright (C) 2002 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.gui.extensions;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.StringSelection;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import weka.classifiers.Classifier;
import weka.classifiers.rules.FR3;
import weka.core.Instance;
import weka.core.Instances;
import weka.gui.ExtensionFileFilter;
import weka.gui.GenericObjectEditor;
import weka.gui.MainMenuExtension;
import weka.gui.PropertyPanel;
import weka.gui.visualize.ClassPanel;

/**
 * PreferenceStructureVisualizer. This class draws the preference structure given
 * a classifier and a query instance.
 * <p>
 *  
 * Date created: 08/07/2008
 * @author Jens Christian Hhn (huehn@gmx.net)
 * @version $Revision: 1.0 $
 * @see JPanel
 */
public class PreferenceStructureVisualizer extends JPanel implements MainMenuExtension {

  /** for serialization */
  private static final long serialVersionUID = -1614936185683699590L;



  /**
   * Inner class to handle rendering the axis
   * 
   * Date created: 08/07/2008
   * @author Jens Christian Hhn (huehn@gmx.net)
   * @version $Revision: 1.0 $
   * @see JPanel
   */
  private class AxisPanel extends JPanel {

    /** for serialization */
    private static final long serialVersionUID = 1136859710812942079L;

    private boolean m_vertical = false;

    private int PAD = 5;

    private FontMetrics m_fontMetrics;

    private int m_fontHeight;

    @Override
    public Rectangle getBounds() {
      return new Rectangle(this.getPreferredSize());
    }

    public AxisPanel(boolean vertical) {
      m_vertical = vertical;
      this.setBackground(Color.white);
      String fontFamily = this.getFont().getFamily();
      Font newFont = new Font(fontFamily, Font.PLAIN, 12);
      this.setFont(newFont);
    }

    public Dimension getPreferredSize() {
      int maxWidth = 0;
      for (int i = 0; m_trainingInstances != null
	  && i < m_trainingInstances.numClasses(); i++) {
	if (maxWidth < m_fontMetrics.stringWidth(m_trainingInstances
	    .classAttribute().value(i))) {
	  maxWidth = m_fontMetrics.stringWidth(m_trainingInstances
	      .classAttribute().value(i));
	}
      }

      if (m_fontMetrics == null) {
	Graphics g = this.getGraphics();
	m_fontMetrics = g.getFontMetrics();
	m_fontHeight = m_fontMetrics.getHeight();
      }
      if (!m_vertical) {
	return new Dimension(this.getSize().width - maxWidth, PAD + 8
	    + m_fontHeight);
      }
      return new Dimension(15 + maxWidth, this.getSize().height);
    }

    public void paintComponent(Graphics g) {
      super.paintComponent(g);
      this.setBackground(Color.white);
      if (m_fontMetrics == null) {
	m_fontMetrics = g.getFontMetrics();
	m_fontHeight = m_fontMetrics.getHeight();
      }
      
      int maxWidth = 0;
      for (int i = 0; m_trainingInstances != null
	  && i < m_trainingInstances.numClasses(); i++) {
	if (maxWidth < m_fontMetrics.stringWidth(m_trainingInstances
	    .classAttribute().value(i))) {
	  maxWidth = m_fontMetrics.stringWidth(m_trainingInstances
	      .classAttribute().value(i));
	}
      }

      Dimension d = this.getSize();
      Dimension d2 = m_boundaryPanel.getSize();
      g.setColor(Color.black);
      int hf = m_fontMetrics.getAscent();
      int stringPad = (d2.width < d2.height ? d2.width : d2.height)
	  / (m_trainingInstances != null ? (m_trainingInstances.numClasses() + (d2.width < d2.height ? 1
	      : 0))
	      : 1);
      if (!m_vertical) {
	 g.drawLine(d.width, d.height-1, d.width-d2.width, d.height-1);
	// try and draw some scale values
	if (getInstances() != null) {
	  if (maxWidth > 8*(int)(d.width / (m_trainingInstances.numClasses() + (d2.width < d2.height ? 1 : 0)))/10-10){
	    maxWidth = 8 * d.width / (m_trainingInstances.numClasses() + (d2.width < d2.height ? 1 : 0)) / 10-10;
	  }

	  for (int i = 0; i < m_trainingInstances.numClasses(); i++) {
	    String classString = m_trainingInstances.classAttribute().value(i);
	    int classStringWith = m_fontMetrics.stringWidth(classString);
	    
	    while (classStringWith > maxWidth){
	      classString = (String) classString.subSequence(0, classString.length()-3);
	      classStringWith = m_fontMetrics.stringWidth(classString);
	    }
	    
	    
	    g.drawString(classString, d.width - d2.width + stringPad * i
		+ stringPad / 2 - classStringWith / 2, PAD + hf + 2);
	  }

	  String scoreString = "Score";
	  int scoreStringWidth = m_fontMetrics.stringWidth(scoreString);

	  g.drawString(scoreString, d.width - d2.width + stringPad
	      * m_trainingInstances.numClasses() + stringPad / 2
	      - scoreStringWidth / 2, PAD + hf + 2);

	}
      } else {
	 g.drawLine(d.width-1, 0, d.width-1, d2.height);
	// try and draw some scale values
	if (getInstances() != null) {

	  for (int i = 0; i < m_trainingInstances.numClasses(); i++) {
	    if (maxWidth < m_fontMetrics.stringWidth(m_trainingInstances
		.classAttribute().value(i))) {
	      maxWidth = m_fontMetrics.stringWidth(m_trainingInstances
		  .classAttribute().value(i));
	    }
	  }

	  for (int i = 0; i < m_trainingInstances.numClasses(); i++) {
	    String classString = m_trainingInstances.classAttribute().value(i);
	    int classStringHeight = m_fontMetrics.getHeight();
	    g.drawString(classString, d.width - PAD - 2
		- m_fontMetrics.stringWidth(classString), PAD + stringPad * i
		+ stringPad / 2 - classStringHeight / 2);
	  }
	}
      }
    }
  }

  /** the number of visualizer windows we have open. */
  protected static int m_WindowCount = 0;

  /** whether the exit if there are no more windows open */
  protected static boolean m_ExitIfNoWindowsOpen = true;

  /** the training instances */
  private Instances m_trainingInstances;

  /** the classifier to use */
  private Classifier m_classifier;

  // plot area dimensions
  protected int m_plotAreaWidth = 512;

  protected int m_plotAreaHeight = 384;

  /** the plotting panel */
  protected PreferenceStructurePanel m_boundaryPanel;
  
  
  // combo boxes for selecting the class attribute, class values (for
  // colouring pixels), and visualization attributes
  protected JComboBox m_classAttBox = new JComboBox();

  protected JComboBox m_LooBox = new JComboBox();

  protected Dimension COMBO_SIZE = new Dimension(
      (int) (m_plotAreaWidth * 0.75), m_classAttBox.getPreferredSize().height);

  protected JButton m_startBut = new JButton("Start");

  protected JButton m_GrayscaleBtn = new JButton("Grayscale Mode");
  
  protected JButton m_copyToClipboardBtn = new JButton("Copy LaTeX to Clipboard");
  
  

  protected JPanel m_controlPanel;

  protected ClassPanel m_classPanel = new ClassPanel();

  // separate panels for rendering axis information
  private AxisPanel m_xAxisPanel;

  private AxisPanel m_yAxisPanel;

  protected GenericObjectEditor m_classifierEditor = new GenericObjectEditor(); // the
										// widget
										// to
										// select
										// the
										// classifier

  protected PropertyPanel m_ClassifierPanel = new PropertyPanel(
      m_classifierEditor);

  /** The file chooser for selecting arff files */
  protected JFileChooser m_FileChooser = new JFileChooser(new File(System
      .getProperty("user.dir")));

  protected ExtensionFileFilter m_arffFileFilter = new ExtensionFileFilter(
      Instances.FILE_EXTENSION, "Arff data files");

  protected JLabel dataFileLabel = new JLabel(); // stores the name of the data
						  // file (currently stores
						  // relation name rather than
						  // filename)

  protected JButton chooseButton = new JButton("Open File"); // button to
							      // choose a data
							      // file

  /**
   * Creates a new <code>PreferenceStructureVisualizer</code> instance.
   */
  public PreferenceStructureVisualizer() {

    setLayout(new BorderLayout());
    m_classAttBox.setMinimumSize(COMBO_SIZE);
    m_classAttBox.setPreferredSize(COMBO_SIZE);
    m_classAttBox.setMaximumSize(COMBO_SIZE);
    m_classAttBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
	if (m_classAttBox.getItemCount() != 0) {
	  try {
	    m_classPanel.setCindex(m_classAttBox.getSelectedIndex());
	    plotTrainingData();
	  } catch (Exception ex) {
	  }

	}
      }
    });

    m_LooBox.setMinimumSize(COMBO_SIZE);
    m_LooBox.setPreferredSize(COMBO_SIZE);
    m_LooBox.setMaximumSize(COMBO_SIZE);

    m_classPanel.setMinimumSize(new Dimension((int) COMBO_SIZE.getWidth() * 2,
	(int) COMBO_SIZE.getHeight() * 2));
    m_classPanel.setPreferredSize(new Dimension(
	(int) COMBO_SIZE.getWidth() * 2, (int) COMBO_SIZE.getHeight() * 2));

    m_controlPanel = new JPanel();
    m_controlPanel.setLayout(new BorderLayout());

    JPanel dataChooseHolder = new JPanel(new BorderLayout());
    dataChooseHolder.setBorder(BorderFactory.createTitledBorder("Dataset"));
    dataChooseHolder.add(dataFileLabel, BorderLayout.WEST);

    m_FileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    m_FileChooser.addChoosableFileFilter(m_arffFileFilter);
    chooseButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
	try {
	  setInstancesFromFileQ();

	  if (m_classifierEditor.getValue() == null)
	    m_classifierEditor.setValue(new FR3());

	  int classIndex = m_classAttBox.getSelectedIndex();
	  if (m_trainingInstances != null && m_classifier != null
	      && (m_trainingInstances.attribute(classIndex).isNominal())) {
	    m_startBut.setEnabled(true);
	    m_copyToClipboardBtn.setEnabled(false);
	    m_GrayscaleBtn.setEnabled(false);
	    plotTrainingData();
	  }

	} catch (Exception ex) {
	  ex.printStackTrace(System.out);
	  System.err.println("exception");
	}

      }
    });
    dataChooseHolder.add(chooseButton, BorderLayout.EAST);

    JPanel classifierHolder = new JPanel();
    classifierHolder.setBorder(BorderFactory.createTitledBorder("Classifier"));
    classifierHolder.setLayout(new BorderLayout());
    m_classifierEditor.setClassType(weka.classifiers.Classifier.class);

    m_classifierEditor.addPropertyChangeListener(new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent evt) {
	m_classifier = (Classifier) m_classifierEditor.getValue();
	try {
	  int classIndex = m_classAttBox.getSelectedIndex();
	  if (m_trainingInstances != null && m_classifier != null
	      && (m_trainingInstances.attribute(classIndex).isNominal())) {
	    m_startBut.setEnabled(true);
	    m_copyToClipboardBtn.setEnabled(false);
	    m_GrayscaleBtn.setEnabled(false);
	  }
	} catch (Exception ex) {
	}
	;
      }
    });
    classifierHolder.add(m_ClassifierPanel, BorderLayout.CENTER);

    JPanel cHolder = new JPanel();
    cHolder.setBorder(BorderFactory.createTitledBorder("Class Attribute"));
    cHolder.add(m_classAttBox);

    JPanel vAttHolder = new JPanel();
    vAttHolder.setLayout(new GridLayout(2, 1));
    vAttHolder.setBorder(BorderFactory
	.createTitledBorder("Visualization Instances"));
    vAttHolder.add(m_LooBox);

    JPanel colOne = new JPanel();
    colOne.setLayout(new BorderLayout());
    colOne.add(dataChooseHolder, BorderLayout.NORTH);
    colOne.add(cHolder, BorderLayout.CENTER);

    JPanel colTwo = new JPanel();
    colTwo.setLayout(new BorderLayout());

    colTwo.add(classifierHolder, BorderLayout.NORTH);
    colTwo.add(vAttHolder, BorderLayout.CENTER);
        
    

    JPanel startPanel = new JPanel();
    startPanel.setBorder(BorderFactory.createTitledBorder("Plotting"));
    startPanel.setLayout(new BorderLayout());
    startPanel.add(m_copyToClipboardBtn, BorderLayout.SOUTH);
    startPanel.add(m_startBut, BorderLayout.CENTER);
    startPanel.add(m_GrayscaleBtn, BorderLayout.NORTH);

    // colTwo.add(startPanel, BorderLayout.SOUTH);

    m_controlPanel.add(colOne, BorderLayout.WEST);
    m_controlPanel.add(colTwo, BorderLayout.CENTER);

    add(m_controlPanel, BorderLayout.NORTH);



    m_boundaryPanel = new PreferenceStructurePanel(m_plotAreaWidth, m_plotAreaHeight);

    JPanel gfxPanel = new JPanel();
    gfxPanel.setLayout(new BorderLayout());
    gfxPanel.setBorder(BorderFactory.createEtchedBorder());

    gfxPanel.add(m_boundaryPanel, BorderLayout.CENTER);
    m_xAxisPanel = new AxisPanel(false);
    gfxPanel.add(m_xAxisPanel, BorderLayout.NORTH);
    m_yAxisPanel = new AxisPanel(true);
    gfxPanel.add(m_yAxisPanel, BorderLayout.WEST);

    JPanel containerPanel = new JPanel();
    containerPanel.setLayout(new BorderLayout());
    containerPanel.add(gfxPanel, BorderLayout.CENTER);
    add(containerPanel, BorderLayout.CENTER);

    JPanel rightHandToolsPanel = new JPanel(); // this panel contains the
						// widgets to the right of the
						// BoundaryPanel.
    rightHandToolsPanel.setLayout(new BoxLayout(rightHandToolsPanel,
	BoxLayout.PAGE_AXIS));

    JButton newWindowButton = new JButton("Open a new window"); // the button
								// for spawning
								// a new window
								// for the
								// program.

    newWindowButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
	try {
	  Instances newTrainingData = null;
	  Classifier newClassifier = null;
	  if (m_trainingInstances != null)
	    newTrainingData = new Instances(m_trainingInstances);
	  if (m_classifier != null)
	    newClassifier = Classifier.makeCopy(m_classifier);
	  createNewVisualizerWindow(newClassifier, newTrainingData);
	} catch (Exception ex) {
	  ex.printStackTrace();
	}
      }
    });
    JPanel newWindowHolder = new JPanel();
    newWindowHolder.add(newWindowButton);
    rightHandToolsPanel.add(newWindowHolder);
    rightHandToolsPanel.add(startPanel);

    containerPanel.add(rightHandToolsPanel, BorderLayout.EAST);
    
    
    m_copyToClipboardBtn.setEnabled(false);
    m_copyToClipboardBtn.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
	if (m_boundaryPanel.m_latexCode != null){
	  try{
	    Clipboard systemClip = Toolkit.getDefaultToolkit().getSystemClipboard();
	    StringSelection  ssel = new StringSelection(m_boundaryPanel.m_latexCode); 
	    systemClip.setContents(ssel,null);
	  }catch(Exception ex){
	    System.err.println("LaTeX-Output: Error whilte writing to clipboard - using console instead");
	    System.out.println(m_boundaryPanel.m_latexCode);
	  }
	}
      }
    });

    m_startBut.setEnabled(false);
    m_startBut.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
	if (m_startBut.getText().equals("Start")) {
	  if (m_trainingInstances != null && m_classifier != null) {
	    try {

	      int BPSuccessCode = setUpBoundaryPanel(); // set up the boundary
							// panel, find out if it
							// was successful or
							// not.

	      if (BPSuccessCode == 1)
		JOptionPane.showMessageDialog(null,
		    "Error: Kernel Bandwidth can't be less than zero!");
	      else if (BPSuccessCode == 2) {
		JOptionPane
		    .showMessageDialog(null,
			"Error: Kernel Bandwidth must be less than the number of training instances!");
	      } else {
		m_boundaryPanel.m_LOOData = new Instances(m_trainingInstances);

		m_boundaryPanel.m_Instance = (Instance) m_boundaryPanel.m_LOOData
		    .instance(m_LooBox.getSelectedIndex()).copy();
		m_boundaryPanel.m_LOOData.delete(m_LooBox.getSelectedIndex());

		m_boundaryPanel.start();
		m_GrayscaleBtn.setEnabled(true);
		m_copyToClipboardBtn.setEnabled(true);
		m_startBut.setText("Stop");
		setControlEnabledStatus(false);
	      }
	    } catch (Exception ex) {
	      ex.printStackTrace();
	    }
	  }
	} else {
	  m_boundaryPanel.stopPlotting();
	  m_startBut.setText("Start");
	  m_GrayscaleBtn.setEnabled(false);
	  m_copyToClipboardBtn.setEnabled(false);
	  setControlEnabledStatus(true);
	}
      }
    });

    m_GrayscaleBtn.setEnabled(false);
    m_GrayscaleBtn.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
	if (m_trainingInstances != null && m_classifier != null
	    && m_boundaryPanel.m_classifierWasBuilt) {
	  if (m_GrayscaleBtn.getText().equals("Grayscale")) {
	    m_boundaryPanel.m_Grayscale = true;
	    m_GrayscaleBtn.setText("Color");
	  } else {
	    m_boundaryPanel.m_Grayscale = false;
	    m_GrayscaleBtn.setText("Grayscale");
	  }
	  m_boundaryPanel.replot();
	}
      }
    });

    m_boundaryPanel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
	m_startBut.setText("Start");
	setControlEnabledStatus(true);
      }
    });

    m_classPanel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
	m_boundaryPanel.replot();
      }
    });

    // set up a mouse listener for the boundary panel.
    m_boundaryPanel.addMouseListener(new MouseAdapter() {
      public void mouseClicked(MouseEvent e) {
	if (m_trainingInstances != null) {
	  if (m_startBut.getText().equals("Stop")) // we are plotting
	    return;

	  try {
	    plotTrainingData();
	  } catch (Exception ex) {
	  }
	  m_xAxisPanel.repaint(0, 0, 0, m_xAxisPanel.getWidth(), m_xAxisPanel
	      .getHeight());
	  m_yAxisPanel.repaint(0, 0, 0, m_yAxisPanel.getWidth(), m_yAxisPanel
	      .getHeight());
	}
      }
    });
  }

  /**
   * Set the enabled status of the controls
   * 
   * @param status
   *                a <code>boolean</code> value
   */
  private void setControlEnabledStatus(boolean status) {
    m_classAttBox.setEnabled(status);
    m_LooBox.setEnabled(status);
    m_FileChooser.setEnabled(status);
    chooseButton.setEnabled(status);
  }

  /**
   * Set a classifier to use
   * 
   * @param newClassifier
   *                the classifier to use
   * @exception Exception
   *                    if an error occurs
   */
  public void setClassifier(Classifier newClassifier) throws Exception {

    m_classifier = newClassifier;

    try {
      int classIndex = m_classAttBox.getSelectedIndex();

      if ((m_classifier != null) && (m_trainingInstances != null)
	  && (m_trainingInstances.attribute(classIndex).isNominal())) {
	m_startBut.setEnabled(true);
      } else
	m_startBut.setEnabled(false);
    } catch (Exception e) {
    }

  }

  /**
   * Sets up the bounds on our x and y axes to fit the dataset. Also repaints
   * the x and y axes.
   */
  private void computeBounds() {
    m_xAxisPanel.repaint(0, 0, 0, m_xAxisPanel.getWidth(), m_xAxisPanel
	.getHeight());
    m_yAxisPanel.repaint(0, 0, 0, m_yAxisPanel.getWidth(), m_yAxisPanel
	.getHeight());
  }

  /**
   * Get the training instances
   * 
   * @return the training instances
   */
  public Instances getInstances() {
    return m_trainingInstances;
  }

  /**
   * Set the training instances
   * 
   * @param inst
   *                the instances to use
   */
  public void setInstances(Instances inst) throws Exception {
    if (inst == null) {
      m_trainingInstances = inst;
      m_classPanel.setInstances(m_trainingInstances);
      return;
    }

    if (inst.numAttributes() < 3) {
      throw new Exception("Not enough attributes in the data to visualize!");
    }
    m_trainingInstances = inst;
    m_classPanel.setInstances(m_trainingInstances);
    // setup combo boxes
    String[] instanceStrings = new String[m_trainingInstances.numAttributes()];
    for (int i = 0; i < m_trainingInstances.numAttributes(); i++) {
      instanceStrings[i] = m_trainingInstances.attribute(i).name();
    }
    final Vector xAttNames = new Vector();
    for (int i = 0; i < m_trainingInstances.numInstances(); i++) {
      xAttNames.addElement((i + 1) + ": "
	  + m_trainingInstances.instance(i).toString());
    }

    m_classAttBox.setModel(new DefaultComboBoxModel(instanceStrings));
    m_LooBox.setModel(new DefaultComboBoxModel(xAttNames));

    m_classAttBox.setSelectedIndex(m_trainingInstances.numAttributes() - 1); // select
									      // last
									      // attribute
									      // as
									      // class
									      // by
									      // default.
									      // -jimmy

    m_classAttBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
	configureForClassAttribute();
      }
    });

    m_LooBox.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent e) {
	if (e.getStateChange() == ItemEvent.SELECTED) {
	  computeBounds();
	  repaint();
	  try {
	    plotTrainingData();
	  } catch (Exception ex) {
	  }
	}
      }
    });

    configureForClassAttribute();

    computeBounds();
    revalidate();
    repaint();

    m_boundaryPanel.m_classifierWasBuilt = false;

  }

  /**
   * Set up the class values combo boxes
   */
  private void configureForClassAttribute() {
    int classIndex = m_classAttBox.getSelectedIndex();
    if (classIndex >= 0) {
      // see if this is a nominal attribute
      if (!m_trainingInstances.attribute(classIndex).isNominal()
	  || m_classifier == null) {
	m_startBut.setEnabled(false);
      } else {
	m_startBut.setEnabled(true);
      }
    }
  }

  /**
   * Queries the user for a file to load instances from, then loads the
   * instances in a background process. This is done in the IO thread, and an
   * error message is popped up if the IO thread is busy.
   */
  public void setInstancesFromFileQ() {

    // if (m_IOThread == null) {
    int returnVal = m_FileChooser.showOpenDialog(this);
    if (returnVal == JFileChooser.APPROVE_OPTION) {
      File selected = m_FileChooser.getSelectedFile();

      try {
	java.io.Reader r = new java.io.BufferedReader(new java.io.FileReader(
	    selected));
	Instances i = new Instances(r);
	setInstances(i);

	// dataFileLabel.setText(selected.getName());
	dataFileLabel.setText(i.relationName());
      } catch (Exception e) {
	JOptionPane.showMessageDialog(this, "Can't load at this time,\n"
	    + "currently busy with other IO", "Load Instances",
	    JOptionPane.WARNING_MESSAGE);
	e.printStackTrace();

      }
    }
  }

  /**
   * Sets up the BoundaryPanel object so that it is ready for plotting.
   * 
   * @return an error code:<br/> 0 - SUCCESS<br/> 1 - ERROR - Kernel bandwidth <
   *         0<br/> 2 - ERROR - Kernel bandwidth >= number of training
   *         instances.
   */
  public int setUpBoundaryPanel() throws Exception {
    int returner = 0; // OK code.
    m_trainingInstances.setClassIndex(m_classAttBox.getSelectedIndex());
    m_boundaryPanel.setClassifier(m_classifier);
    m_boundaryPanel.setTrainingData(m_trainingInstances);

    return returner;
  }

  /**
   * Plots the training data on-screen. Also does all of the setup required for
   * this to work.
   */
  public void plotTrainingData() throws Exception {
    m_boundaryPanel.initialize();
    setUpBoundaryPanel();
    computeBounds();
    m_boundaryPanel.plotTrainingData();
  }

  /**
   * Stops the plotting thread.
   */
  public void stopPlotting() {
    m_boundaryPanel.stopPlotting();
  }

  /**
   * Sets whether System.exit gets called when no more windows are open.
   * 
   * @param value
   *                if TRUE then a System.exit call is ossued after the last
   *                window gets closed.
   */
  public static void setExitIfNoWindowsOpen(boolean value) {
    m_ExitIfNoWindowsOpen = value;
  }

  /**
   * Gets whether System.exit gets called after the last window gets closed
   * 
   * @return TRUE if System.exit gets called after last window got closed.
   */
  public static boolean getExitIfNoWindowsOpen() {
    return m_ExitIfNoWindowsOpen;
  }

  /**
   * Creates a new GUI window with all of the PreferenceStructureVisualizer trappings,
   * 
   * @param classifier
   *                The classifier to use in the new window. May be null.
   * @param instances
   *                The dataset to visualize on in the new window. May be null.
   */
  public static void createNewVisualizerWindow(Classifier classifier,
      Instances instances) throws Exception {
    m_WindowCount++;

    final javax.swing.JFrame jf = new javax.swing.JFrame(
	"Preference Structure Visualizer");
    jf.getContentPane().setLayout(new BorderLayout());
    final PreferenceStructureVisualizer bv = new PreferenceStructureVisualizer();
    jf.getContentPane().add(bv, BorderLayout.CENTER);
    jf.setSize(bv.getMinimumSize());
    jf.addWindowListener(new java.awt.event.WindowAdapter() {
      public void windowClosing(java.awt.event.WindowEvent e) {
	m_WindowCount--;
	bv.stopPlotting();
	jf.dispose();
	if ((m_WindowCount == 0) && m_ExitIfNoWindowsOpen) {
	  System.exit(0);
	}
      }
    });

    jf.pack();
    jf.setVisible(true);
    jf.setResizable(false);

    if (classifier == null)
      bv.setClassifier(null);
    else {
      bv.setClassifier(classifier);
      bv.m_classifierEditor.setValue(classifier);
    }

    if (instances == null)
      bv.setInstances(null);
    else {
      bv.setInstances(instances);

      try {
	bv.dataFileLabel.setText(instances.relationName());
	bv.plotTrainingData();
	bv.m_classPanel.setCindex(bv.m_classAttBox.getSelectedIndex());
	bv.repaint(0, 0, 0, bv.getWidth(), bv.getHeight());
      } catch (Exception ex) {
      }
    }

  }

  public void fillFrame(Component frame) {
    m_WindowCount++;

    if (frame instanceof JFrame) {
      final javax.swing.JFrame jf = (JFrame) frame;

      jf.getContentPane().setLayout(new BorderLayout());
      final PreferenceStructureVisualizer bv = new PreferenceStructureVisualizer();
      jf.getContentPane().add(bv, BorderLayout.CENTER);
      jf.setSize(bv.getMinimumSize());
      jf.addWindowListener(new java.awt.event.WindowAdapter() {
	public void windowClosing(java.awt.event.WindowEvent e) {
	  m_WindowCount--;
	  bv.stopPlotting();
	  jf.dispose();
	  if ((m_WindowCount == 0) && m_ExitIfNoWindowsOpen) {
	    System.exit(0);
	  }
	}
      });

      jf.pack();
      jf.setVisible(true);
      jf.setResizable(false);

    } else if (frame instanceof JInternalFrame) {
      final javax.swing.JInternalFrame jf = (JInternalFrame) frame;

      jf.getContentPane().setLayout(new BorderLayout());
      final PreferenceStructureVisualizer bv = new PreferenceStructureVisualizer();
      jf.getContentPane().add(bv, BorderLayout.CENTER);
      jf.setSize(bv.getMinimumSize());

      jf.pack();

    }

  }

  public ActionListener getActionListener(JFrame owner) {
    return null;
  }

  public String getMenuTitle() {
    return "Preference Structure Visualizer";
  }

  public String getSubmenuTitle() {
    return null;
  }
}
