/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    FR3.java
 *    Copyright (C) 2008 Jens Christian Hhn
 *
 *    (based upon)  MultiClassClassifier.java
 *    Copyright (C) 1999 Eibe Frank,Len Trigg
 *
 */

package weka.classifiers.rules;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Enumeration;
import java.util.Random;
import java.util.Vector;
import weka.classifiers.Classifier;
import weka.classifiers.PreferenceStructureHandler;
import weka.classifiers.RandomizableSingleClassifierEnhancer;
import weka.core.Attribute;
import weka.core.Capabilities;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.Range;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformationHandler;
import weka.core.Utils;
import weka.core.WeightedInstancesHandler;
import weka.core.Capabilities.Capability;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.MakeIndicator;
import weka.filters.unsupervised.instance.RemoveWithValues;


/**
 <!-- globalinfo-start -->
 * <em>FR3 - Fuzzy Round Robin Ripper </em><br>
 * The FR3 algorithm is a pairwise classification algorithm which uses a modified version of the Ripper algorithm in combination with fuzzification techniques. The classification itself is achieved via pairwise preferences instead of voting.  
 * <p/>
 <!-- globalinfo-end -->
 *
 * <!-- technical-bibtex-start -->
 * BibTeX:
 * <pre>
 *  &#64;article{huehn2008,
 *    author = {Jens Christian Hhn and Eyke Hllermeier},
 *    journal = {},
 *    title = {FR3: A Fuzzy Rule Learner for Inducing Reliable Classifiers},
 *    year = {2008}
 * }
 * </pre>
 * <p/>
 <!-- technical-bibtex-end -->
 *
 <!-- options-start -->
 * Valid options are: <p/>
 * 
 *    
 * <pre> -d
 *  If set, the classifier will safe the binary models to the
 *  specified path after learning. When the first classification 
 *  is done, the models will be retrieved.</pre>
 *  
 *  <pre> -P &lt;str&gt;
 *  The path to which the models are saved.
 *  (default "./")</pre>
 *     
 * <pre> -B
 *  If set, the classifier learns all n(n-1) models, otherwise it
 *  learns only one class per pair and thus n(n-1)/2 models</pre>
 * 
 * <pre> -S &lt;num&gt;
 *  Random number seed.
 *  (default 1)</pre>
 * 
 * <pre> -D
 *  If set, classifier is run in debug mode and
 *  may output additional info to the console</pre>
 * 
 * <pre> -W
 *  Full name of base classifier.
 *  (default: weka.classifiers.rules.FRip)</pre>
 * 
 * <pre> 
 * Options specific to classifier weka.classifiers.rules.FRip:
 * </pre>
 * <pre> -F &lt;number of folds&gt;
 *  Set number of folds for REP
 *  One fold is used as pruning set.
 *  (default 3)</pre>
 * 
 * <pre> -N &lt;min. weights&gt;
 *  Set the minimal weights of instances
 *  within a split.
 *  (default 2.0)</pre>
 * 
 * <pre> -O &lt;number of runs&gt;
 *  Set the number of runs of
 *  optimizations. (Default: 2)</pre>
 * 
 * <pre> -D
 *  Set whether turn on the
 *  debug mode (Default: false)</pre>
 * 
 * <pre> -S &lt;seed&gt;
 *  The seed of randomization
 *  (Default: 1)</pre>
 * 
 * <pre> -E
 *  Whether NOT check the error rate&gt;=0.5
 *  in stopping criteria  (default: check)</pre>
 * 
 * 
 <!-- options-end -->
 * Date created: 08/07/2008
 * @author Jens Christian Hhn (huehn@gmx.net)
 * @version $Revision: 1.0 $
 */
public class FR3 
extends RandomizableSingleClassifierEnhancer 
implements OptionHandler,TechnicalInformationHandler,WeightedInstancesHandler, PreferenceStructureHandler {

  /** for serialization */
  private static final long serialVersionUID = 3493596602554738869L;

  /** The classifiers. */
  public Classifier [] m_Classifiers;

  /** The filters used to transform the class. */
  protected Filter[] m_ClassFilters;

  /** Internal copy of the class attribute for output purposes */
  protected Attribute m_ClassAttribute;

  /** A transformed dataset header used by the  1-against-1 method */
  protected Instances m_TwoClassDataset;

  /** Switch which determines whether all n(n-1) models are learned or only n(n-1)/2 models.*/
  protected boolean m_learnBothModelsForEachPairOfClasses = false;

  /** An array which stores the class distribution of the training data*/
  private double[] aprioriClassDistribution;

  /** A switch which determines whether the binary models should be cached to the harddrive (to free RAM for other classifiers temporarily)*/
  private boolean m_safeModelToDiskAfterLearning = false;

  /** The path to which the binary models are written */
  private String m_modelSafePath = "./" ;

  /** A flag which shows that the saved models may have already been loaded to the RAM*/ 
  private boolean m_filesWereLoaded = false;

  /** The filename which is used for pairwise models which are written to disk.*/
  private String m_filenamePattern = "";

  /** The random variable generator*/
  private Random m_Random = null;


  /**
   * Constructor.
   */
  public FR3(){
    m_Classifier = new weka.classifiers.rules.FRIP();
    
    // initialize the underlying FRipper algorithm as suggested
    ((FRIP)m_Classifier).setCheckErrorRate(false);
    ((FRIP)m_Classifier).setMinNo(2);
  }

  /**
   * String describing default classifier.
   * 
   * @return the default classifier classname
   */
  protected String defaultClassifierString() {
    return "weka.classifiers.rules.FRip";
  }


  /**
   * Returns default capabilities of the classifier.
   *
   * @return      the capabilities of this classifier
   */
  public Capabilities getCapabilities() {
    Capabilities result = super.getCapabilities();

    // class
    result.disableAllClasses();
    result.disableAllClassDependencies();
    result.disable(Capability.BINARY_CLASS);
    result.enable(Capability.NOMINAL_CLASS);


    return result;
  }

  /**
   * Returns an instance of a TechnicalInformation object, containing 
   * detailed information about the technical background of this class,
   * e.g., paper reference or book this class is based on.
   * 
   * @return the technical information about this class
   */
  public TechnicalInformation getTechnicalInformation() {
    TechnicalInformation 	result;

    result = new TechnicalInformation(Type.ARTICLE);
    result.setValue(Field.AUTHOR, "Jens Christian Hhn and Eyke Hllermeier");
    result.setValue(Field.TITLE, "FR3: A Fuzzy Rule Learner for Inducing Reliable Classifiers");
    result.setValue(Field.YEAR, "2008");
    result.setValue(Field.JOURNAL, "");

    return result;
  }

  /**
   * Builds the classifiers.
   *
   * @param instances the training data.
   * @throws Exception if a classifier can't be built
   */
  public void buildClassifier(Instances instances) throws Exception {
    Instances newInsts;

    // can classifier handle the data?
    getCapabilities().testWithFail(instances);

    // remove instances with missing class
    instances = new Instances(instances);
    instances.deleteWithMissingClass();

    m_Random = instances.getRandomNumberGenerator(m_Seed); 



    m_filenamePattern = "FR3_" + instances.relationName() + "_" + System.currentTimeMillis();

    // Find the edges of the known dataspace
    double[][] dataspaceEdges = new double[instances.numAttributes()-1][2];
    for (int i = 0; i < instances.numAttributes(); i++){
      Instances dataWithoutMissingValues = new Instances(instances);
      dataWithoutMissingValues.deleteWithMissing(i);
      if (instances.attribute(i).isNumeric()){
	if (dataWithoutMissingValues.numInstances() == 1){
	  dataspaceEdges[i][0] = dataWithoutMissingValues.kthSmallestValue(i, 1) ;
	  dataspaceEdges[i][1] = dataWithoutMissingValues.kthSmallestValue(i, 1) ;
	}else if (dataWithoutMissingValues.numInstances() > 1){
	  dataspaceEdges[i][0] = dataWithoutMissingValues.kthSmallestValue(i, 1) ;
	  dataspaceEdges[i][1] = dataWithoutMissingValues.kthSmallestValue(i, dataWithoutMissingValues.numInstances());
	}
      }
    }
    // Inform FRipper about the dataspace edges
    if (m_Classifier instanceof FRIP) {
      ((FRIP)m_Classifier).setDataspaceEdges(dataspaceEdges);
      ((FRIP)m_Classifier).m_dataAllClasses = new Instances(instances);
    }


    // Retrieve the a priori class distribution
    aprioriClassDistribution = new double[instances.numClasses()];
    for (int i = 0; i < instances.numInstances();i++){
      aprioriClassDistribution[(int)instances.instance(i).classValue()]+=1.0/instances.numInstances();
    }





    if (m_Classifier == null) {
      throw new Exception("No base classifier has been set!");
    }

    m_TwoClassDataset = null;




    // generate fastvector of pairs
    FastVector pairs = new FastVector();
    for (int i=0; i<instances.numClasses(); i++) {
      for (int j=0; j<instances.numClasses(); j++) {
	if (j==i) continue;
	if (j<i && !m_learnBothModelsForEachPairOfClasses) continue;
	int[] pair = new int[2];
	pair[0] = i; pair[1] = j;
	pairs.addElement(pair);
      }
    }

    int numClassifiers  = pairs.size();
    m_Classifiers = Classifier.makeCopies(m_Classifier, numClassifiers);
    m_ClassFilters = new Filter[numClassifiers];

    // generate the classifiers
    for (int i=0; i<numClassifiers; i++) {
      RemoveWithValues classFilter = new RemoveWithValues();
      classFilter.setAttributeIndex("" + (instances.classIndex() + 1));
      classFilter.setModifyHeader(true);
      classFilter.setInvertSelection(true);
      classFilter.setNominalIndicesArr((int[])pairs.elementAt(i));
      Instances tempInstances = new Instances(instances, 0);
      tempInstances.setClassIndex(-1);
      classFilter.setInputFormat(tempInstances);
      newInsts = Filter.useFilter(instances, classFilter);
      if (newInsts.numInstances() > 0) {
	newInsts.setClassIndex(instances.classIndex());
	m_Classifiers[i].buildClassifier(newInsts);
	if (m_safeModelToDiskAfterLearning){
	  saveModel(m_Classifiers[i], m_filenamePattern + i);
	  m_Classifiers[i] = null;
	}
	m_ClassFilters[i] = classFilter;
      } else {
	m_Classifiers[i] = null;
	m_ClassFilters[i] = null;
      }
    }

    // construct a two-class header version of the dataset
    m_TwoClassDataset = new Instances(instances, 0);
    int classIndex = m_TwoClassDataset.classIndex();
    m_TwoClassDataset.setClassIndex(-1);
    m_TwoClassDataset.deleteAttributeAt(classIndex);
    FastVector classLabels = new FastVector();
    classLabels.addElement("class0");
    classLabels.addElement("class1");
    m_TwoClassDataset.insertAttributeAt(new Attribute("class", classLabels),
	classIndex);
    m_TwoClassDataset.setClassIndex(classIndex);

    m_ClassAttribute = instances.classAttribute();
  }


  public double[] distributionForInstance(Instance inst) throws Exception {
    /* When the models were saved to the disk, they are retrieved when the first
     * classification is done. Afterwards they stay in the main memory.
     */
    if (m_safeModelToDiskAfterLearning && !m_filesWereLoaded){
      for (int i = 0; i < m_Classifiers.length; i++){
	if (m_ClassFilters != null && m_ClassFilters[i] != null && m_Classifiers[i]==null)
	  m_Classifiers[i] = loadModel(m_filenamePattern + i);
      }
      m_filesWereLoaded = true;
    }

    // retrieve the P,C,I matrices
    double[][][] relationSet = relationsForInstance(inst);
    double[][] preferenceRelation = relationSet[0];
    double[][] conflictRelation = relationSet[1];
    double[][] ignoranceRelation = relationSet[2];
    
    // calculate the class scores
    double[] classScores = new double[preferenceRelation.length];
    for (int i = 0; i < preferenceRelation.length; i++){
      for (int j = 0; j < preferenceRelation[i].length; j++){
	if (i!=j){
	  classScores[i] +=
	    preferenceRelation[i][j] + 
	    conflictRelation[i][j]  * 0.5 +
	    ignoranceRelation[i][j]  * 
	    (aprioriClassDistribution[i] / (aprioriClassDistribution[i]+aprioriClassDistribution[j]))
	    ;

	}
      }
      if (Double.isNaN(classScores[i])) classScores[i] = 0;
    }
    

   
    if (Utils.sum(classScores)==0)
      return classScores;

    Utils.normalize(classScores);
    return classScores;
  };

  /**
   * Classifies an instance.
   *
   * @param inst the instance to classify
   * @return the classification
   * @throws Exception if instance can't be classified successfully
   */
  public double classifyInstance(Instance inst) throws Exception {
    double[] classScores = distributionForInstance(inst);


    /*
     * Tie-breaking step 1: Find out which classes gain the maximum score 
     */
    double maxValue = classScores[Utils.maxIndex(classScores)];
    double[] ties = new double[classScores.length];
    for (int i = 0; i < classScores.length; i++){
      if (classScores[i] == maxValue){
	ties[i] = aprioriClassDistribution[i];
      }
    }

    classScores = new double[classScores.length];
    classScores[Utils.maxIndex(ties)] = 1;



    /*
     * Tie-breaking step 2: Check whether the tying classes have the same a priori
     * class probability and count these classes. 
     */
    int tieValues = 0;
    maxValue = ties[Utils.maxIndex(ties)];
    for (int i = 0; i < ties.length; i++){
      if (ties[i] == maxValue){
	tieValues++;
      }
    }

    /*
     * Tie-breaking step 3: If the tying classes have the same a priori probabilities,
     * then use randomization to determine the winner among these classes
     */
    if (tieValues > 1){
      tieValues=0;
      maxValue = ties[Utils.maxIndex(ties)];
      int[] stillTying = new int[ties.length];

      for (int i = 0; i < classScores.length; i++){
	if (ties[i] == maxValue){
	  stillTying[tieValues] = i;
	  tieValues++;
	}
      }
      return stillTying[m_Random.nextInt(tieValues)];
    }

    return Utils.maxIndex(classScores);
  }

  /**
   * Retrieves the preference, conflict and ignorance matrices for a single instance
   * @param inst The instance for which the PCI-matrices shall be created
   * @return [0][][] preference matrix, [1][][] conflict matrix, [2][][] ignorance matrix
   * @throws Exception
   */
  public double[][][] relationsForInstance(Instance inst) throws Exception {
    double[][] prefRelation = new double[m_ClassAttribute.numValues()][m_ClassAttribute.numValues()];
    double[][] conflictRelation = new double[m_ClassAttribute.numValues()][m_ClassAttribute.numValues()];
    double[][] ignoranceRelation = new double[m_ClassAttribute.numValues()][m_ClassAttribute.numValues()];


    for(int i = 0; i < m_Classifiers.length; i++) {
      if (m_Classifiers[i] != null) {
	Instance tempInst = (Instance)inst.copy(); 
	tempInst.setDataset(m_TwoClassDataset);

	// Copy the result to another array, in case the result of the classifier was indeed only binary
	double [] current = new double[4];
	double [] result = m_Classifiers[i].distributionForInstance(tempInst);
	for (int j = 0; j < result.length; j++){
	  current[j] = result[j];
	}


	int[] pair = null;
	if (m_Classifiers.length>1){
	  Range range = new Range(((RemoveWithValues)m_ClassFilters[i])
	      .getNominalIndices());
	  range.setUpper(m_ClassAttribute.numValues());
	  pair = range.getSelection();
	}else{
	  pair = new int[]{0,1};
	}

	prefRelation[pair[0]][pair[1]] += current[0];
	prefRelation[pair[1]][pair[0]] += current[1];
	conflictRelation[pair[0]][pair[1]] += current[2];
	conflictRelation[pair[1]][pair[0]] += current[2];
	ignoranceRelation[pair[0]][pair[1]] += current[3];
	ignoranceRelation[pair[1]][pair[0]] += current[3];

      }
    }


    return new double[][][] {prefRelation, conflictRelation, ignoranceRelation};
  }


  /**
   * Prints the classifiers.
   * 
   * @return a string representation of the classifier
   */
  public String toString() {

    if (m_Classifiers == null) {
      return "FR3: No model built yet.";
    }
    StringBuffer text = new StringBuffer();
    text.append("FR3\n\n");
    for (int i = 0; i < m_Classifiers.length; i++) {
      text.append("Classifier ").append(i + 1);
      if (m_Classifiers[i] != null) {
	if ((m_ClassFilters != null) && (m_ClassFilters[i] != null)) {
	  if (m_ClassFilters[i] instanceof RemoveWithValues) {
	    Range range = new Range(((RemoveWithValues)m_ClassFilters[i])
		.getNominalIndices());
	    range.setUpper(m_ClassAttribute.numValues());
	    int[] pair = range.getSelection();
	    text.append(", " + (pair[0]+1) + " vs " + (pair[1]+1));
	  } else if (m_ClassFilters[i] instanceof MakeIndicator) {
	    text.append(", using indicator values: ");
	    text.append(((MakeIndicator)m_ClassFilters[i]).getValueRange());
	  }
	}
	text.append('\n');
	text.append(m_Classifiers[i].toString() + "\n\n");
      } else {
	text.append(" Skipped (no training examples)\n");
      }
    }

    return text.toString();
  }

  /**
   * Returns an enumeration describing the available options
   *
   * @return an enumeration of all the available options
   */
  public Enumeration listOptions()  {

    Vector vec = new Vector(3);


    vec.addElement(new Option(
	"\tSets whether the binary models are saved to disk after learning before classification",
	"d", 0, "-d"));
    vec.addElement(new Option(
	"\tSets the path to which the binary models are saved (if enabled)",
	"P", 1, "-P <string>"));
    vec.addElement(new Option(
	"\tSets whether for each pair one (resulting n(n-1)/2) classifier is learned or both (resulting in n(n-1)) classifiers",
	"B", 0, "-B"));

    Enumeration enu = super.listOptions();
    while (enu.hasMoreElements()) {
      vec.addElement(enu.nextElement());
    }
    return vec.elements();
  }

  /**
   * Parses a given list of options. <p/>
   *
   <!-- options-start -->
   * Valid options are: <p/>
   *  
   * <pre> -d
   *  If set, the classifier will safe the binary models to the
   *  specified path after learning. When the first classification 
   *  is done, the models will be retrieved.</pre>
   *  
   *  <pre> -P &lt;str&gt;
   *  The path to which the models are saved.
   *  (default "./")</pre>
   *     
   * <pre> -B
   *  If set, the classifier learns all n(n-1) models, otherwise it
   *  learns only one class per pair and thus n(n-1)/2 models</pre>
   *  
   * <pre> -S &lt;num&gt;
   *  Random number seed.
   *  (default 1)</pre>
   * 
   * <pre> -D
   *  If set, classifier is run in debug mode and
   *  may output additional info to the console</pre>
   * 
   * <pre> -W
   *  Full name of base classifier.
   *  (default: weka.classifiers.rules.FRip)</pre>
   * 
   * <pre> 
   * Options specific to classifier weka.classifiers.rules.FRip:
   * </pre>
   * <pre> -F &lt;number of folds&gt;
   *  Set number of folds for REP
   *  One fold is used as pruning set.
   *  (default 3)</pre>
   * 
   * <pre> -N &lt;min. weights&gt;
   *  Set the minimal weights of instances
   *  within a split.
   *  (default 2.0)</pre>
   * 
   * <pre> -O &lt;number of runs&gt;
   *  Set the number of runs of
   *  optimizations. (Default: 2)</pre>
   * 
   * <pre> -D
   *  Set whether turn on the
   *  debug mode (Default: false)</pre>
   * 
   * <pre> -S &lt;seed&gt;
   *  The seed of randomization
   *  (Default: 1)</pre>
   * 
   * <pre> -E
   *  Whether NOT check the error rate&gt;=0.5
   *  in stopping criteria  (default: check)</pre>
   *   
   <!-- options-end -->
   *
   * @param options the list of options as an array of strings
   * @throws Exception if an option is not supported
   */
  public void setOptions(String[] options) throws Exception {
    m_learnBothModelsForEachPairOfClasses = Utils.getFlag('B', options);
    m_safeModelToDiskAfterLearning = Utils.getFlag('d', options);

    m_modelSafePath = Utils.getOption('P', options);


    String seedString = Utils.getOption('S', options);
    if (seedString.length() != 0)
      m_Seed = (int)Long.parseLong(seedString);
    else 
      m_Seed = 1;

    super.setOptions(options);
  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] superOptions = super.getOptions();
    String [] options = new String [superOptions.length+4];

    int current = 0;

    if(m_learnBothModelsForEachPairOfClasses)
      options[current++] = "-B";

    if(m_safeModelToDiskAfterLearning)
      options[current++] = "-d";

    options[current++] = "-P"; options[current++] = "" + m_modelSafePath;



    System.arraycopy(superOptions, 0, options, current, 
	superOptions.length);

    current += superOptions.length;
    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }

  public String safeModelToDiskAfterLearningTipText() {
    return "Sets whether the pairwise models should be safed to the harddrive after learning and be loaded upon classification";
  }

  /**
   * Sets whether the pairwise models should be safed to the harddrive after learning and be loaded upon classification 
   * @param b true if the models shall be saved to disk
   */

  public void setSafeModelToDiskAfterLearning(boolean b){
    m_safeModelToDiskAfterLearning = b;
  }

  /**
   * Gets whether the pairwise models should be safed to the harddrive after learning and be loaded upon classification 
   * @return true if the models shall be saved to disk
   */
  public boolean getSafeModelToDiskAfterLearning(){
    return m_safeModelToDiskAfterLearning;
  } 

  public String learnBothModelsForEachPairOfClassesTipText() {
    return "Sets whether all n(n-1) models (true) or only the n(n-1)/2 models, which is one model per pair.";
  }


  /**
   * Gets whether all n(n-1) models (true) or only the n(n-1)/2 models, which is one model per pair.
   * @return true for two models per pair, false for only one model for each pair 
   */
  public boolean getLearnBothModelsForEachPairOfClasses(){
    return m_learnBothModelsForEachPairOfClasses;
  }

  /**
   * Sets whether all n(n-1) models (true) or only the n(n-1)/2 models, which is one model per pair.
   * @param b true for two models per pair, false for only one model for each pair 
   */
  public void setLearnBothModelsForEachPairOfClasses(boolean b){
    m_learnBothModelsForEachPairOfClasses=b;
  }

  public String modelSafePathTipText() {
    return "Sets the path to which the pairwise models are temporarily saved, if the saving is activated.";
  }

  /**
   * Set the path to which the models are stored between learning and classification
   * @param p the path to which the models are strored
   */
  public void setModelSafePath(String p){
    m_modelSafePath = p;
  }

  /**
   * Get the path to which the models are stored between learning and classification
   * @return the path to which the models are strored
   */
  public String getModelSafePath(){
    return m_modelSafePath;
  } 


  /**
   * Returns a string describing classifier
   * @return a description suitable for
   * displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return "The FR3 is the Fuzzy Round Robin Classifier. It learns pairwise classifiers " +
    "on a multi-class dataset. The proposed base classifier FRipper (FRip) is " +
    "a fuzzification of Cohen's Ripper (JRip). The classification decision is achieved " +
    "with the help of a preference, conflict and ignorance relation instead of voting.";
  }

  /**
   * A function which loads a saved model from disk to memory. After loading the model, the
   * file will be deleted.
   * @param fileName The filename for the file in which the classifier is stored
   * @return The loaded classifier
   * @throws FileNotFoundException
   * @throws IOException
   * @throws ClassNotFoundException
   */
  private Classifier loadModel(String fileName) throws FileNotFoundException, IOException, ClassNotFoundException{
    try{
      FileInputStream fis = new FileInputStream(m_modelSafePath + fileName);
      ObjectInputStream objIn = new ObjectInputStream(new BufferedInputStream(fis));
      Classifier c = (Classifier)objIn.readObject();
      objIn.close();
      fis.close();
      File f = new File(m_modelSafePath + fileName);
      f.delete();
      return c;
    }catch(Exception e){
      System.out.println(m_TwoClassDataset.toSummaryString());
      e.printStackTrace();
    }
    return null;
  }

  /**
   * A function which saves a given model to disk
   * @param c The classifier which shall be saved 
   * @param fileName The filename which describes the file in which the classifier is saved
   * @throws FileNotFoundException
   * @throws IOException
   */
  public void saveModel(Classifier c, String fileName) throws FileNotFoundException, IOException{
    FileOutputStream fos = new FileOutputStream(m_modelSafePath + fileName);
    ObjectOutputStream objOut = new ObjectOutputStream(new BufferedOutputStream(fos));
    objOut.writeObject(c); 
    objOut.close();
    fos.close();
  }

  /**
   * Main method for testing this class.
   *
   * @param argv the options
   * @throws Exception 
   */
  public static void main(String [] argv) throws Exception {
    runClassifier(new FR3(), argv); 
  }

  public String getRevision() {
    return "1.0";
  }  
}
